# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"OpenStackClient plugin for Placement service"

import logging

from osc_lib import utils
from osc_placement import version


LOG = logging.getLogger(__name__)

API_NAME = 'placement'
API_VERSION_OPTION = 'os_placement_api_version'
API_VERSIONS = {v: 'osc_placement.http.SessionClient'
                for v in version.SUPPORTED_VERSIONS}


def make_client(instance):
    client_class = utils.get_client_class(
        API_NAME,
        instance._api_version[API_NAME],
        API_VERSIONS
    )

    ks_filter = {'service_type': API_NAME,
                 'region_name': instance._region_name,
                 'interface': instance.interface}

    LOG.debug('Instantiating placement client: %s', client_class)
    return client_class(session=instance.session,
                        ks_filter=ks_filter,
                        api_version=instance._api_version[API_NAME])


def build_option_parser(parser):
    default = version.NEGOTIATE_VERSIONS[0]
    parser.add_argument(
        '--os-placement-api-version',
        metavar='<placement-api-version>',
        default=utils.env(
            'OS_PLACEMENT_API_VERSION',
            default=default
        ),
        help='Placement API version, default=%s '
             '(Env: OS_PLACEMENT_API_VERSION)' % default)

    return parser



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
import copy
import uuid

from osc_placement.tests.functional import base


class TestInventory(base.BaseTestCase):
    def setUp(self):
        super(TestInventory, self).setUp()

        self.rp = self.resource_provider_create()

    def test_inventory_show(self):
        rp_uuid = self.rp['uuid']
        updates = {
            'min_unit': 1,
            'max_unit': 12,
            'reserved': 0,
            'step_size': 1,
            'total': 12,
            'allocation_ratio': 16.0,
        }
        expected = updates.copy()
        expected['used'] = 0
        args = ['VCPU:%s=%s' % (k, v) for k, v in updates.items()]

        self.resource_inventory_set(rp_uuid, *args)

        self.assertEqual(
            expected,
            self.resource_inventory_show(rp_uuid, 'VCPU', include_used=True),
        )

    def test_inventory_show_not_found(self):
        rp_uuid = self.rp['uuid']

        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_show,
                                rp_uuid, 'VCPU')
        self.assertIn('No inventory of class VCPU for {}'.format(rp_uuid),
                      str(exc))

    def test_inventory_list(self):
        rp_uuid = self.rp['uuid']
        updates = {
            'min_unit': 1,
            'max_unit': 12,
            'reserved': 0,
            'step_size': 1,
            'total': 12,
            'allocation_ratio': 16.0,
        }
        expected = [updates.copy()]
        expected[0]['resource_class'] = 'VCPU'
        expected[0]['used'] = 0
        args = ['VCPU:%s=%s' % (k, v) for k, v in updates.items()]

        self.resource_inventory_set(rp_uuid, *args)

        self.assertEqual(
            expected, self.resource_inventory_list(rp_uuid, include_used=True),
        )

    def test_inventory_delete(self):
        rp_uuid = self.rp['uuid']

        self.resource_inventory_set(rp_uuid, 'VCPU=8')

        self.resource_inventory_delete(rp_uuid, 'VCPU')
        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_show,
                                rp_uuid, 'VCPU')
        self.assertIn('No inventory of class VCPU for {}'.format(rp_uuid),
                      str(exc))

    def test_inventory_delete_not_found(self):
        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_delete,
                                self.rp['uuid'], 'VCPU')
        self.assertIn('No inventory of class VCPU found for delete',
                      str(exc))

    def test_delete_all_inventories(self):
        # Negative test to assert command failure because
        # microversion < 1.5 and --resource-class is not specified.
        self.assertCommandFailed(
            base.ARGUMENTS_REQUIRED % '--resource-class',
            self.resource_inventory_delete,
            'fake_uuid')


class TestSetInventory(base.BaseTestCase):
    def test_fail_if_no_rp(self):
        exc = self.assertRaises(
            base.CommandException,
            self.openstack, 'resource provider inventory set')
        self.assertIn(base.ARGUMENTS_MISSING, str(exc))

    def test_set_empty_inventories(self):
        rp = self.resource_provider_create()
        self.assertEqual([], self.resource_inventory_set(rp['uuid']))

    def test_fail_if_incorrect_resource(self):
        rp = self.resource_provider_create()
        # wrong format
        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_set,
                                rp['uuid'], 'VCPU')
        self.assertIn('must have "name=value"', str(exc))
        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_set,
                                rp['uuid'], 'VCPU==')
        self.assertIn('must have "name=value"', str(exc))
        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_set,
                                rp['uuid'], '=10')
        self.assertIn('must be not empty', str(exc))
        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_set,
                                rp['uuid'], 'v=')
        self.assertIn('must be not empty', str(exc))

        # unknown class
        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_set,
                                rp['uuid'], 'UNKNOWN_CPU=16')
        self.assertIn('Unknown resource class', str(exc))
        # unknown property
        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_set,
                                rp['uuid'], 'VCPU:fake=16')
        self.assertIn('Unknown inventory field', str(exc))

    def test_set_multiple_classes(self):
        rp = self.resource_provider_create()
        resp = self.resource_inventory_set(
            rp['uuid'],
            'VCPU=8',
            'VCPU:max_unit=4',
            'MEMORY_MB=1024',
            'MEMORY_MB:reserved=256',
            'DISK_GB=16',
            'DISK_GB:allocation_ratio=1.5',
            'DISK_GB:min_unit=2',
            'DISK_GB:step_size=2')

        def check(inventories):
            self.assertEqual(8, inventories['VCPU']['total'])
            self.assertEqual(4, inventories['VCPU']['max_unit'])
            self.assertEqual(1024, inventories['MEMORY_MB']['total'])
            self.assertEqual(256, inventories['MEMORY_MB']['reserved'])
            self.assertEqual(16, inventories['DISK_GB']['total'])
            self.assertEqual(2, inventories['DISK_GB']['min_unit'])
            self.assertEqual(2, inventories['DISK_GB']['step_size'])
            self.assertEqual(1.5, inventories['DISK_GB']['allocation_ratio'])

        check({r['resource_class']: r for r in resp})
        resp = self.resource_inventory_list(rp['uuid'])
        check({r['resource_class']: r for r in resp})

    def test_set_known_and_unknown_class(self):
        rp = self.resource_provider_create()
        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_set,
                                rp['uuid'], 'VCPU=8', 'UNKNOWN=4')
        self.assertIn('Unknown resource class', str(exc))
        self.assertEqual([], self.resource_inventory_list(rp['uuid']))

    def test_replace_previous_values(self):
        """Test each new set call replaces previous inventories totally."""
        rp = self.resource_provider_create()
        # set disk inventory first
        self.resource_inventory_set(rp['uuid'], 'DISK_GB=16')
        # set memory and vcpu inventories
        self.resource_inventory_set(rp['uuid'], 'MEMORY_MB=16', 'VCPU=32')
        resp = self.resource_inventory_list(rp['uuid'])
        inv = {r['resource_class']: r for r in resp}
        # no disk inventory as it was overwritten
        self.assertNotIn('DISK_GB', inv)
        self.assertIn('VCPU', inv)
        self.assertIn('MEMORY_MB', inv)

    def test_delete_via_set(self):
        rp = self.resource_provider_create()
        self.resource_inventory_set(rp['uuid'], 'DISK_GB=16')
        self.resource_inventory_set(rp['uuid'])
        self.assertEqual([], self.resource_inventory_list(rp['uuid']))

    def test_fail_if_incorrect_parameters_set_class_inventory(self):
        exc = self.assertRaises(
            base.CommandException,
            self.openstack, 'resource provider inventory class set')
        self.assertIn(base.ARGUMENTS_MISSING, str(exc))
        exc = self.assertRaises(
            base.CommandException,
            self.openstack, 'resource provider inventory class set fake_uuid')
        self.assertIn(base.ARGUMENTS_MISSING, str(exc))
        exc = self.assertRaises(
            base.CommandException,
            self.openstack,
            ('resource provider inventory class set '
             'fake_uuid fake_class --total 5 --unknown 1'))
        self.assertIn('unrecognized arguments', str(exc))
        # Valid RP UUID and resource class, but no inventory field.
        rp = self.resource_provider_create()
        exc = self.assertRaises(
            base.CommandException, self.openstack,
            'resource provider inventory class set %s VCPU' % rp['uuid'])
        self.assertIn(base.ARGUMENTS_REQUIRED % '--total',
                      str(exc))

    def test_set_inventory_for_resource_class(self):
        rp = self.resource_provider_create()
        self.resource_inventory_set(rp['uuid'], 'MEMORY_MB=16', 'VCPU=32')
        self.resource_inventory_class_set(
            rp['uuid'], 'MEMORY_MB', total=128, step_size=16)
        resp = self.resource_inventory_list(rp['uuid'])
        inv = {r['resource_class']: r for r in resp}
        self.assertEqual(128, inv['MEMORY_MB']['total'])
        self.assertEqual(16, inv['MEMORY_MB']['step_size'])
        self.assertEqual(32, inv['VCPU']['total'])

    def test_fail_aggregate_arg_version_handling(self):
        agg = str(uuid.uuid4())
        self.assertCommandFailed(
            'Operation or argument is not supported with version 1.0; '
            'requires at least version 1.3',
            self.resource_inventory_set,
            agg, 'MEMORY_MB=16', aggregate=True)

    def test_amend_multiple_classes(self):
        # Tests that amending previously nonexistent resource class inventories
        # results in creation of inventory for them
        # Create a resource provider with no inventory
        rp = self.resource_provider_create()
        resp = self.resource_inventory_set(
            rp['uuid'],
            'VCPU=8',
            'VCPU:max_unit=4',
            'MEMORY_MB=1024',
            'MEMORY_MB:reserved=256',
            'DISK_GB=16',
            'DISK_GB:allocation_ratio=1.5',
            'DISK_GB:min_unit=2',
            'DISK_GB:step_size=2',
            amend=True)

        def check(inventories):
            self.assertEqual(8, inventories['VCPU']['total'])
            self.assertEqual(4, inventories['VCPU']['max_unit'])
            self.assertEqual(1024, inventories['MEMORY_MB']['total'])
            self.assertEqual(256, inventories['MEMORY_MB']['reserved'])
            self.assertEqual(16, inventories['DISK_GB']['total'])
            self.assertEqual(1.5, inventories['DISK_GB']['allocation_ratio'])
            self.assertEqual(2, inventories['DISK_GB']['min_unit'])
            self.assertEqual(2, inventories['DISK_GB']['step_size'])

        inventories = {r['resource_class']: r for r in resp}
        check(inventories)
        resp = self.resource_inventory_list(rp['uuid'])
        inventories = {r['resource_class']: r for r in resp}
        check(inventories)

        # Test amending of one resource class inventory
        resp = self.resource_inventory_set(
            rp['uuid'],
            'VCPU:allocation_ratio=5.0',
            amend=True)
        inventories = {r['resource_class']: r for r in resp}
        check(inventories)
        self.assertEqual(5.0, inventories['VCPU']['allocation_ratio'])
        resp = self.resource_inventory_list(rp['uuid'])
        inventories = {r['resource_class']: r for r in resp}
        check(inventories)
        self.assertEqual(5.0, inventories['VCPU']['allocation_ratio'])

    def test_dry_run(self):
        rp = self.resource_provider_create()
        resp = self.resource_inventory_set(
            rp['uuid'],
            'VCPU=8',
            'VCPU:max_unit=4',
            'MEMORY_MB=1024',
            'MEMORY_MB:reserved=256',
            'DISK_GB=16',
            'DISK_GB:allocation_ratio=1.5',
            'DISK_GB:min_unit=2',
            'DISK_GB:step_size=2',
            dry_run=True)

        def check(inventories):
            self.assertEqual(8, inventories['VCPU']['total'])
            self.assertEqual(4, inventories['VCPU']['max_unit'])
            self.assertEqual(1024, inventories['MEMORY_MB']['total'])
            self.assertEqual(256, inventories['MEMORY_MB']['reserved'])
            self.assertEqual(16, inventories['DISK_GB']['total'])
            self.assertEqual(2, inventories['DISK_GB']['min_unit'])
            self.assertEqual(2, inventories['DISK_GB']['step_size'])
            self.assertEqual(1.5, inventories['DISK_GB']['allocation_ratio'])

        # We expect the return value from the set command to reflect the values
        # passed to the command (a preview of what would be set if not for
        # --dry-run)
        check({r['resource_class']: r for r in resp})
        # But we expect the return value from the list command to be empty
        # since we used --dry-run and didn't actually effect any changes
        resp = self.resource_inventory_list(rp['uuid'])
        self.assertEqual([], resp)


class TestInventory15(TestInventory):
    VERSION = '1.5'

    def test_delete_all_inventories(self):
        rp = self.resource_provider_create()
        self.resource_inventory_set(rp['uuid'], 'MEMORY_MB=16', 'VCPU=32')
        self.resource_inventory_delete(rp['uuid'])
        self.assertEqual([], self.resource_inventory_list(rp['uuid']))


class TestAggregateInventory(base.BaseTestCase):
    VERSION = '1.3'

    def _test_dry_run(self, agg, rps, old_inventories, amend=False):
        new_resources = ['VCPU:allocation_ratio=5.0',
                         'MEMORY_MB:allocation_ratio=6.0',
                         'DISK_GB:allocation_ratio=7.0']
        resp = self.resource_inventory_set(
            agg, *new_resources, aggregate=True, amend=amend, dry_run=True)
        # Use empty dict to get expected values for full replacement
        inventories = old_inventories if amend else [{}] * len(old_inventories)
        # A list of dict keyed by resource class of dict of inventories
        # Each list element corresponds to one resource provider
        new_inventories = self._get_expected_inventories(
            inventories, new_resources)
        # To compare with actual result, reformat new_inventories to
        # a dict, keyed by resource provider uuid, of dict of inventories
        # keyed by resource class
        expected = {}
        for rp, inventory in zip(rps, new_inventories):
            for rc, inv in inventory.items():
                inv['resource_provider'] = rp['uuid']
                # For full replacement (not amend) case these values should
                # be set to empty string.
                for key in ('max_unit', 'min_unit', 'reserved', 'step_size',
                            'total', 'reserved', 'step_size'):
                    if key not in inv:
                        inv[key] = ''
            expected[rp['uuid']] = inventory
        # Reformat raw response (list of inventories) as well
        resp_dict = collections.defaultdict(dict)
        for row in resp:
            resp_dict[row['resource_provider']][row['resource_class']] = row
        self.assertEqual(expected, resp_dict)
        # Verify the inventories weren't changed (--dry-run)
        for i, rp in enumerate(rps):
            resp = self.resource_inventory_list(rp['uuid'])
            self.assertDictEqual(old_inventories[i],
                                 {r['resource_class']: r for r in resp})

    def _get_expected_inventories(self, old_inventories, resources):
        new_inventories = []
        for old_inventory in old_inventories:
            new_inventory = collections.defaultdict(dict)
            new_inventory.update(copy.deepcopy(old_inventory))
            for resource in resources:
                rc, keyval = resource.split(':')
                key, val = keyval.split('=')
                # Handle allocation ratio which is a float
                val = float(val) if '.' in val else int(val)
                new_inventory[rc][key] = val
                # The resource_class field is added by the osc_placement CLI,
                # so add it to our expected inventories
                if 'resource_class' not in new_inventory[rc]:
                    new_inventory[rc]['resource_class'] = rc
            new_inventories.append(new_inventory)
        return new_inventories

    def _setup_two_resource_providers_in_aggregate(self):
        rps = []
        invs = []
        inventory2 = ['VCPU=8',
                      'VCPU:max_unit=4',
                      'VCPU:allocation_ratio=16.0',
                      'MEMORY_MB=1024',
                      'MEMORY_MB:reserved=256',
                      'MEMORY_MB:allocation_ratio=2.5',
                      'DISK_GB=16',
                      'DISK_GB:allocation_ratio=1.5',
                      'DISK_GB:min_unit=2',
                      'DISK_GB:step_size=2']
        inventory1 = inventory2 + ['VGPU=8',
                                   'VGPU:allocation_ratio=1.0',
                                   'VGPU:min_unit=2',
                                   'VGPU:step_size=2']
        for i, inventory in enumerate([inventory1, inventory2]):
            rps.append(self.resource_provider_create())
            resp = self.resource_inventory_set(rps[i]['uuid'], *inventory)
            # Verify the resource_provider column is not present without
            # --aggregate
            self.assertNotIn('resource_provider', resp)
            invs.append({r['resource_class']: r for r in resp})
        # Put both resource providers in the same aggregate
        agg = str(uuid.uuid4())
        for rp in rps:
            self.resource_provider_aggregate_set(rp['uuid'], agg)
        return rps, agg, invs

    def test_fail_if_no_rps_in_aggregate(self):
        nonexistent_agg = str(uuid.uuid4())
        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_set,
                                nonexistent_agg,
                                'VCPU=8',
                                aggregate=True)
        self.assertIn('No resource providers found in aggregate with uuid {}'
                      .format(nonexistent_agg), str(exc))

    def test_with_aggregate_one_fails(self):
        # Set up some existing inventories with two resource providers
        rps, agg, _invs = self._setup_two_resource_providers_in_aggregate()
        # Set a custom resource class inventory on the first resource provider
        self.resource_class_create('CUSTOM_FOO')
        rp1_uuid = rps[0]['uuid']
        rp1_inv = self.resource_inventory_set(rp1_uuid, 'CUSTOM_FOO=1')
        # Create an allocation for custom resource class on first provider
        consumer = str(uuid.uuid4())
        alloc = 'rp=%s,CUSTOM_FOO=1' % rp1_uuid
        self.resource_allocation_set(consumer, [alloc])
        # Try to set allocation ratio for an aggregate. The first set should
        # fail because we're not going to set the custom resource class (which
        # is equivalent to trying to remove it) and removal isn't allowed if
        # there is an allocation of it present. The second set should succeed
        new_resources = ['VCPU:allocation_ratio=5.0', 'VCPU:total=8']
        exc = self.assertRaises(base.CommandException,
                                self.resource_inventory_set,
                                agg, *new_resources, aggregate=True)
        self.assertIn('Failed to set inventory for 1 of 2 resource providers.',
                      str(exc))
        output = self.output.getvalue() + self.error.getvalue()
        self.assertIn('Failed to set inventory for resource provider %s:' %
                      rp1_uuid, output)
        err_txt = ("Inventory for 'CUSTOM_FOO' on resource provider '%s' in "
                   "use." % rp1_uuid)
        self.assertIn(err_txt, output)
        # Placement will default the following internally
        placement_defaults = ['VCPU:max_unit=2147483647',
                              'VCPU:min_unit=1',
                              'VCPU:reserved=0',
                              'VCPU:step_size=1']
        # Get expected inventory for the second resource provider (succeeded)
        new_inventories = self._get_expected_inventories(
            # Since inventories are expected to be fully replaced,
            # use empty dict for old inventory
            [{}],
            new_resources + placement_defaults)
        resp = self.resource_inventory_list(rps[1]['uuid'])
        self.assertDictEqual(new_inventories[0],
                             {r['resource_class']: r for r in resp})
        # First resource provider should have remained the same (failed)
        resp = self.resource_inventory_list(rp1_uuid)
        self.assertDictEqual({r['resource_class']: r for r in rp1_inv},
                             {r['resource_class']: r for r in resp})

    def _test_with_aggregate(self, amend=False):
        # Set up some existing inventories with two resource providers
        rps, agg, old_invs = self._setup_two_resource_providers_in_aggregate()
        # Verify that --dry-run works properly
        self._test_dry_run(agg, rps, old_invs, amend=amend)
        # If we're not amending inventory, set some defaults that placement
        # will set internally and return when we list inventories later
        if not amend:
            old_invs = []
            defaults = {'max_unit': 2147483647,
                        'min_unit': 1,
                        'reserved': 0,
                        'step_size': 1}
            default_inventory = {'VCPU': copy.deepcopy(defaults)}
            for rp in rps:
                old_invs.append(default_inventory)
        # Now, go ahead and update an allocation ratio and verify
        new_resources = ['VCPU:allocation_ratio=5.0',
                         'VCPU:total=8']
        resp = self.resource_inventory_set(agg, *new_resources, aggregate=True,
                                           amend=amend)
        # Verify the resource_provider column is present with --aggregate
        for rp in resp:
            self.assertIn('resource_provider', rp)
        new_inventories = self._get_expected_inventories(old_invs,
                                                         new_resources)
        for i, rp in enumerate(rps):
            resp = self.resource_inventory_list(rp['uuid'])
            self.assertDictEqual(new_inventories[i],
                                 {r['resource_class']: r for r in resp})

    def test_with_aggregate(self):
        self._test_with_aggregate()

    def test_amend_with_aggregate(self):
        self._test_with_aggregate(amend=True)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse

from osc_lib.command import command
from osc_lib import utils

from osc_placement.resources import common
from osc_placement import version


BASE_URL = '/resource_providers'
ALLOCATIONS_URL = BASE_URL + '/{uuid}/allocations'


class CreateResourceProvider(command.ShowOne, version.CheckerMixin):
    """Create a new resource provider"""

    def get_parser(self, prog_name):
        parser = super(CreateResourceProvider, self).get_parser(prog_name)

        parser.add_argument(
            '--parent-provider',
            metavar='<parent_provider>',
            help='UUID of the parent provider.'
                 ' Omit for no parent.'
                 ' This option requires at least'
                 ' ``--os-placement-api-version 1.14``.'
        )
        parser.add_argument(
            '--uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Name of the resource provider'
        )

        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        data = {'name': parsed_args.name}

        if 'uuid' in parsed_args and parsed_args.uuid:
            data['uuid'] = parsed_args.uuid
        if ('parent_provider' in parsed_args
                and parsed_args.parent_provider):
            self.check_version(version.ge('1.14'))
            data['parent_provider_uuid'] = parsed_args.parent_provider

        resp = http.request('POST', BASE_URL, json=data)
        resource = http.request('GET', resp.headers['Location']).json()

        fields = ('uuid', 'name', 'generation')
        if self.compare_version(version.ge('1.14')):
            fields += ('root_provider_uuid', 'parent_provider_uuid')

        return fields, utils.get_dict_properties(resource, fields)


class ListResourceProvider(command.Lister, version.CheckerMixin):
    """List resource providers"""

    def get_parser(self, prog_name):
        parser = super(ListResourceProvider, self).get_parser(prog_name)

        parser.add_argument(
            '--uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='Name of the resource provider'
        )
        parser.add_argument(
            '--resource',
            metavar='<resource_class>=<value>',
            default=[],
            action='append',
            help='A resource class value pair indicating an '
                 'amount of resource of a specified class that a provider '
                 'must have the capacity to serve. May be repeated.\n\n'
                 'This param requires at least '
                 '``--os-placement-api-version 1.4``.'
        )
        parser.add_argument(
            '--in-tree',
            metavar='<in_tree>',
            help='Restrict listing to the same "provider tree"'
                 ' as the specified provider UUID.'
                 ' This option requires at least'
                 ' ``--os-placement-api-version 1.14``.'
        )
        parser.add_argument(
            '--required',
            metavar='<required>',
            action='append',
            default=[],
            help='A required trait. May be repeated. Resource providers '
                 'must collectively contain all of the required traits. '
                 'This option requires at least '
                 '``--os-placement-api-version 1.18``. '
                 'Since ``--os-placement-api-version 1.39`` the value of '
                 'this parameter can be a comma separated list of trait names '
                 'to express OR relationship between those traits.'
        )
        parser.add_argument(
            '--forbidden',
            metavar='<forbidden>',
            action='append',
            default=[],
            help='A forbidden trait. May be repeated. Returned resource '
                 'providers must not contain any of the specified traits. '
                 'This option requires at least '
                 '``--os-placement-api-version 1.22``.'
        )
        # NOTE(tetsuro): --aggregate-uuid is deprecated in Jan 2020 in 1.x
        # release. Do not remove before Jan 2021 and a 2.x release.
        aggregate_group = parser.add_mutually_exclusive_group()
        aggregate_group.add_argument(
            "--member-of",
            default=[],
            action='append',
            metavar='<member_of>',
            help='A list of comma-separated UUIDs of the resource provider '
                 'aggregates. The returned resource providers must be '
                 'associated with at least one of the aggregates identified '
                 'by uuid. This param requires at least '
                 '``--os-placement-api-version 1.3`` and can be repeated to '
                 'add(restrict) the condition with '
                 '``--os-placement-api-version 1.24`` or greater. '
                 'For example, to get candidates either in agg1 or in agg2 '
                 'and definitely in agg3, specify:\n\n'
                 '``--member_of <agg1>,<agg2> --member_of <agg3>``'
        )
        aggregate_group.add_argument(
            '--aggregate-uuid',
            default=[],
            action='append',
            metavar='<aggregate_uuid>',
            help=argparse.SUPPRESS
        )

        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        filters = {}
        if parsed_args.name:
            filters['name'] = parsed_args.name
        if parsed_args.uuid:
            filters['uuid'] = parsed_args.uuid
        if parsed_args.aggregate_uuid:
            self.check_version(version.ge('1.3'))
            self.deprecated_option_warning("--aggregate-uuid", "--member-of")
            filters['member_of'] = 'in:' + ','.join(parsed_args.aggregate_uuid)
        if parsed_args.resource:
            self.check_version(version.ge('1.4'))
            filters['resources'] = ','.join(
                resource.replace('=', ':')
                for resource in parsed_args.resource)
        if 'in_tree' in parsed_args and parsed_args.in_tree:
            self.check_version(version.ge('1.14'))
            filters['in_tree'] = parsed_args.in_tree

        # We need to handle required and forbidden together as they all end up
        # in the same query param on the API.
        # First just check that the requested feature is aligned with the
        # request microversion
        required_traits = []
        if 'required' in parsed_args and parsed_args.required:
            self.check_version(version.ge('1.18'))
            if any(',' in required for required in parsed_args.required):
                self.check_version(version.ge('1.39'))
            required_traits = parsed_args.required

        forbidden_traits = []
        if 'forbidden' in parsed_args and parsed_args.forbidden:
            self.check_version(version.ge('1.22'))
            forbidden_traits = ['!' + f for f in parsed_args.forbidden]

        # Then collect the required query params containing both required and
        # forbidden traits
        filters['required'] = common.get_required_query_param_from_args(
            required_traits, forbidden_traits)

        if 'member_of' in parsed_args and parsed_args.member_of:
            # Fail if --member-of but not high enough microversion.
            self.check_version(version.ge('1.3'))
            filters['member_of'] = [
                'in:' + aggs for aggs in parsed_args.member_of]

        resources = http.request(
            'GET', BASE_URL, params=filters).json()['resource_providers']

        fields = ('uuid', 'name', 'generation')
        if self.compare_version(version.ge('1.14')):
            fields += ('root_provider_uuid', 'parent_provider_uuid')

        rows = (utils.get_dict_properties(r, fields) for r in resources)
        return fields, rows


class ShowResourceProvider(command.ShowOne, version.CheckerMixin):
    """Show resource provider details"""

    def get_parser(self, prog_name):
        parser = super(ShowResourceProvider, self).get_parser(prog_name)
        # TODO(avolkov): show by uuid or name
        parser.add_argument(
            '--allocations',
            action='store_true',
            help='include the info on allocations of the provider resources'
        )
        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )

        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = BASE_URL + '/' + parsed_args.uuid
        resource = http.request('GET', url).json()

        fields = ('uuid', 'name', 'generation')
        if self.compare_version(version.ge('1.14')):
            fields += ('root_provider_uuid', 'parent_provider_uuid')

        if parsed_args.allocations:
            allocs_url = ALLOCATIONS_URL.format(uuid=parsed_args.uuid)
            allocs = http.request('GET', allocs_url).json()['allocations']
            resource['allocations'] = allocs
            fields += ('allocations',)

        return fields, utils.get_dict_properties(resource, fields)


class SetResourceProvider(command.ShowOne, version.CheckerMixin):
    """Update an existing resource provider"""

    def get_parser(self, prog_name):
        parser = super(SetResourceProvider, self).get_parser(prog_name)

        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help='A new name of the resource provider',
            required=True
        )
        parser.add_argument(
            '--parent-provider',
            metavar='<parent_provider>',
            help='UUID of the parent provider.'
                 ' Can only be set if the resource provider has no parent yet.'
                 ' This option requires at least'
                 ' ``--os-placement-api-version 1.14``.'
        )

        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = BASE_URL + '/' + parsed_args.uuid
        data = dict(name=parsed_args.name)
        # Not knowing the previous state of a resource the client cannot catch
        # it, but if the user tries to re-parent a resource provider the server
        # returns an easy to understand error:
        #     Unable to save resource provider RP-ID:
        #     Object action update failed because:
        #     re-parenting a provider is not currently allowed.
        #     (HTTP 400)
        if ('parent_provider' in parsed_args
                and parsed_args.parent_provider):
            self.check_version(version.ge('1.14'))
            data['parent_provider_uuid'] = parsed_args.parent_provider
        resource = http.request('PUT', url, json=data).json()

        fields = ('uuid', 'name', 'generation')
        if self.compare_version(version.ge('1.14')):
            fields += ('root_provider_uuid', 'parent_provider_uuid')

        return fields, utils.get_dict_properties(resource, fields)


class DeleteResourceProvider(command.Command):
    """Delete a resource provider"""

    def get_parser(self, prog_name):
        parser = super(DeleteResourceProvider, self).get_parser(prog_name)

        # TODO(avolkov): delete by uuid or name
        parser.add_argument(
            'uuid',
            metavar='<uuid>',
            help='UUID of the resource provider'
        )

        return parser

    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = BASE_URL + '/' + parsed_args.uuid
        http.request('DELETE', url)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import json
import logging
import random

import fixtures

from openstackclient import shell
from oslotest import base
from placement.tests.functional.fixtures import capture
from placement.tests.functional.fixtures import placement


# A list of logger names that will be reset to a log level
# of WARNING. Due (we think) to a poor interaction between the
# way osc does logging and oslo.logging, all packages are producing
# DEBUG logs. This results in test attachments (when capturing logs)
# that are sometimes larger than subunit.parser can deal with. The
# packages chosen here are ones that do not provide useful information.
RESET_LOGGING = [
    'keystoneauth.session',
    'oslo_policy.policy',
    'placement.objects.trait',
    'placement.objects.resource_class',
    'placement.objects.resource_provider',
    'oslo_concurrency.lockutils',
    'osc_lib.shell',
]

RP_PREFIX = 'osc-placement-functional-tests-'

ARGUMENTS_MISSING = 'the following arguments are required'
ARGUMENTS_REQUIRED = 'the following arguments are required: %s'


class CommandException(Exception):
    def __init__(self, *args, **kwargs):
        super(CommandException, self).__init__(args[0])
        self.cmd = kwargs['cmd']


class BaseTestCase(base.BaseTestCase):
    VERSION = '1.0'

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.useFixture(capture.Logging())
        self.placement = self.useFixture(placement.PlacementFixture())

        # Work around needing to reset the session's notion of where
        # we are going.
        def mock_get(obj, instance, owner):
            return obj.factory(instance)

        # NOTE(cdent): This is fragile, but is necessary to work around
        # the rather complex start up optimizations that are done in osc_lib.
        # If/when osc_lib changes this will at least fail fast.
        self.useFixture(fixtures.MonkeyPatch(
            'osc_lib.clientmanager.ClientCache.__get__',
            mock_get))

        # Reset log level on a set of packages. See comment on RESET_LOGGING
        # assigment, above.
        for name in RESET_LOGGING:
            logging.getLogger(name).setLevel(logging.WARNING)

    def openstack(self, cmd, may_fail=False, use_json=False,
                  may_print_to_stderr=False):
        to_exec = []
        # Make all requests as a noauth admin user.
        to_exec += [
            '--os-endpoint', self.placement.endpoint,
            '--os-token', self.placement.token,
            '--os-auth-type', 'admin_token',
        ]
        if self.VERSION is not None:
            to_exec += ['--os-placement-api-version', self.VERSION]
        to_exec += cmd.split()
        if use_json:
            to_exec += ['-f', 'json']

        # Context manager here instead of setUp because we only want
        # output trapping around the run().
        self.output = io.StringIO()
        self.error = io.StringIO()
        stdout_fix = fixtures.MonkeyPatch('sys.stdout', self.output)
        stderr_fix = fixtures.MonkeyPatch('sys.stderr', self.error)
        with stdout_fix, stderr_fix:
            try:
                os_shell = shell.OpenStackShell()
                return_code = os_shell.run(to_exec)
            # Catch SystemExit to trap some error responses, mostly from the
            # argparse lib which has a tendency to exit for you instead of
            # politely telling you it wants to.
            except SystemExit as exc:
                return_code = exc.code

        # We may have error/warning messages in stderr, so treat it
        # separately from the stdout.
        output = self.output.getvalue()
        error = self.error.getvalue()

        if return_code:
            msg = 'Command: "%s"\noutput: %s' % (' '.join(to_exec), error)
            if not may_fail:
                raise CommandException(msg, cmd=' '.join(to_exec))

        if use_json and output:
            output = json.loads(output)

        if may_print_to_stderr:
            return output, error

        if error:
            msg = ('Test code error - The command did not fail but it '
                   'has a warning message. Set the "may_print_to_stderr" '
                   'argument to true to get and validate the message:\n'
                   'Command: "%s"\nstderr: %s') % (
                ' '.join(to_exec), error)
            raise CommandException(msg, cmd=' '.join(to_exec))
        return output

    def rand_name(self, name='', prefix=None):
        """Generate a random name that includes a random number

        :param str name: The name that you want to include
        :param str prefix: The prefix that you want to include
        :return: a random name. The format is
                 '<prefix>-<name>-<random number>'.
                 (e.g. 'prefixfoo-namebar-154876201')
        :rtype: string
        """
        # NOTE(lajos katona): This method originally is in tempest-lib.
        randbits = str(random.randint(1, 0x7fffffff))
        rand_name = randbits
        if name:
            rand_name = name + '-' + rand_name
        if prefix:
            rand_name = prefix + '-' + rand_name
        return rand_name

    def assertCommandFailed(self, message, func, *args, **kwargs):
        signature = [func]
        signature.extend(args)
        try:
            func(*args, **kwargs)
            self.fail('Command does not fail as required (%s)' % signature)
        except CommandException as e:
            self.assertIn(
                message, str(e),
                'Command "%s" fails with different message' % e.cmd)

    def resource_provider_create(self,
                                 name='',
                                 parent_provider_uuid=None):
        if not name:
            name = self.rand_name(name='', prefix=RP_PREFIX)

        to_exec = 'resource provider create ' + name
        if parent_provider_uuid is not None:
            to_exec += ' --parent-provider ' + parent_provider_uuid
        res = self.openstack(to_exec, use_json=True)

        def cleanup():
            try:
                self.resource_provider_delete(res['uuid'])
            except CommandException as exc:
                # may have already been deleted by a test case
                err_message = str(exc).lower()
                if 'no resource provider' not in err_message:
                    raise
        self.addCleanup(cleanup)

        return res

    def resource_provider_set(self, uuid, name, parent_provider_uuid=None):
        to_exec = 'resource provider set ' + uuid + ' --name ' + name
        if parent_provider_uuid is not None:
            to_exec += ' --parent-provider ' + parent_provider_uuid
        return self.openstack(to_exec, use_json=True)

    def resource_provider_show(self, uuid, allocations=False):
        cmd = 'resource provider show ' + uuid
        if allocations:
            cmd = cmd + ' --allocations'

        return self.openstack(cmd, use_json=True)

    def resource_provider_list(self, uuid=None, name=None,
                               aggregate_uuids=None, resources=None,
                               in_tree=None, required=None, forbidden=None,
                               member_of=None, may_print_to_stderr=False):
        to_exec = 'resource provider list'
        if uuid:
            to_exec += ' --uuid ' + uuid
        if name:
            to_exec += ' --name ' + name
        if aggregate_uuids:
            to_exec += ' ' + ' '.join(
                '--aggregate-uuid %s' % a for a in aggregate_uuids)
        if resources:
            to_exec += ' ' + ' '.join('--resource %s' % r for r in resources)
        if in_tree:
            to_exec += ' --in-tree ' + in_tree
        if required:
            to_exec += ' ' + ' '.join('--required %s' % t for t in required)
        if forbidden:
            to_exec += ' ' + ' '.join('--forbidden %s' % f for f in forbidden)
        if member_of:
            to_exec += ' ' + ' '.join(
                ['--member-of %s' % m for m in member_of])

        return self.openstack(
            to_exec, use_json=True, may_print_to_stderr=may_print_to_stderr)

    def resource_provider_delete(self, uuid):
        return self.openstack('resource provider delete ' + uuid)

    def resource_allocation_show(self, consumer_uuid, columns=()):
        cmd = 'resource provider allocation show ' + consumer_uuid
        cmd += ' '.join(' --column %s' % c for c in columns)
        return self.openstack(cmd, use_json=True)

    def resource_allocation_set(self, consumer_uuid, allocations,
                                project_id=None, user_id=None,
                                consumer_type=None, use_json=True,
                                may_print_to_stderr=False):
        cmd = 'resource provider allocation set {allocs} {uuid}'.format(
            uuid=consumer_uuid,
            allocs=' '.join('--allocation {}'.format(a) for a in allocations)
        )
        if project_id:
            cmd += ' --project-id %s' % project_id
        if user_id:
            cmd += ' --user-id %s' % user_id
        if consumer_type:
            cmd += ' --consumer-type %s' % consumer_type
        result = self.openstack(
            cmd, use_json=use_json, may_print_to_stderr=may_print_to_stderr)

        def cleanup(uuid):
            try:
                self.openstack('resource provider allocation delete ' + uuid)
            except CommandException as exc:
                # may have already been deleted by a test case
                if 'not found' in str(exc).lower():
                    pass
        self.addCleanup(cleanup, consumer_uuid)

        return result

    def resource_allocation_unset(
        self, consumer_uuid, provider=None, resource_class=None, use_json=True,
        columns=(),
    ):
        cmd = 'resource provider allocation unset %s' % consumer_uuid
        if resource_class:
            cmd += ' ' + ' '.join(
                '--resource-class %s' % rc for rc in resource_class)
        if provider:
            # --provider can be specified multiple times so if we only get
            # a single string value convert to a list.
            if isinstance(provider, str):
                provider = [provider]
            cmd += ' ' + ' '.join(
                '--provider %s' % rp_uuid for rp_uuid in provider)

        cmd += ' '.join(' --column %s' % c for c in columns)

        result = self.openstack(cmd, use_json=use_json)

        def cleanup(uuid):
            try:
                self.openstack('resource provider allocation delete ' + uuid)
            except CommandException as exc:
                # may have already been deleted by a test case
                if 'not found' in str(exc).lower():
                    pass
        self.addCleanup(cleanup, consumer_uuid)

        return result

    def resource_allocation_delete(self, consumer_uuid):
        cmd = 'resource provider allocation delete ' + consumer_uuid
        return self.openstack(cmd)

    def resource_inventory_show(
        self, uuid, resource_class, *, include_used=False,
    ):
        resource = self.openstack(
            f'resource provider inventory show {uuid} {resource_class}',
            use_json=True,
        )
        if not include_used:
            del resource['used']

        return resource

    def resource_inventory_list(self, uuid, *, include_used=False):
        resources = self.openstack(
            f'resource provider inventory list {uuid}',
            use_json=True,
        )
        if not include_used:
            for resource in resources:
                del resource['used']

        return resources

    def resource_inventory_delete(self, uuid, resource_class=None):
        cmd = 'resource provider inventory delete {uuid}'.format(uuid=uuid)
        if resource_class:
            cmd += ' --resource-class ' + resource_class
        self.openstack(cmd)

    def resource_inventory_set(self, uuid, *resources, **kwargs):
        opts = []
        if kwargs.get('aggregate'):
            opts.append('--aggregate')
        if kwargs.get('amend'):
            opts.append('--amend')
        if kwargs.get('dry_run'):
            opts.append('--dry-run')
        fmt = 'resource provider inventory set {uuid} {resources} {opts}'
        cmd = fmt.format(
            uuid=uuid,
            resources=' '.join(['--resource %s' % r for r in resources]),
            opts=' '.join(opts))
        return self.openstack(cmd, use_json=True)

    def resource_inventory_class_set(self, uuid, resource_class, **kwargs):
        opts = ['--%s=%s' % (k, v) for k, v in kwargs.items()]
        cmd = 'resource provider inventory class set {uuid} {rc} {opts}'.\
            format(uuid=uuid, rc=resource_class, opts=' '.join(opts))
        return self.openstack(cmd, use_json=True)

    def resource_provider_show_usage(self, uuid):
        return self.openstack('resource provider usage show ' + uuid,
                              use_json=True)

    def resource_show_usage(self, project_id, user_id=None):
        cmd = 'resource usage show %s' % project_id
        if user_id:
            cmd += ' --user-id %s' % user_id
        return self.openstack(cmd, use_json=True)

    def resource_provider_aggregate_list(self, uuid):
        return self.openstack('resource provider aggregate list ' + uuid,
                              use_json=True)

    def resource_provider_aggregate_set(self, uuid, *aggregates,
                                        **kwargs):
        generation = kwargs.get('generation')
        cmd = 'resource provider aggregate set %s ' % uuid
        cmd += ' '.join('--aggregate %s' % aggregate
                        for aggregate in aggregates)
        if generation is not None:
            cmd += ' --generation %s' % generation
        return self.openstack(cmd, use_json=True)

    def resource_class_list(self):
        return self.openstack('resource class list', use_json=True)

    def resource_class_show(self, name):
        return self.openstack('resource class show ' + name, use_json=True)

    def resource_class_create(self, name):
        return self.openstack('resource class create ' + name)

    def resource_class_set(self, name):
        return self.openstack('resource class set ' + name)

    def resource_class_delete(self, name):
        return self.openstack('resource class delete ' + name)

    def trait_list(self, name=None, associated=False):
        cmd = 'trait list'
        if name:
            cmd += ' --name ' + name
        if associated:
            cmd += ' --associated'
        return self.openstack(cmd, use_json=True)

    def trait_show(self, name):
        cmd = 'trait show %s' % name
        return self.openstack(cmd, use_json=True)

    def trait_create(self, name):
        cmd = 'trait create %s' % name
        self.openstack(cmd)

        def cleanup():
            try:
                self.trait_delete(name)
            except CommandException as exc:
                # may have already been deleted by a test case
                err_message = str(exc).lower()
                if 'http 404' not in err_message:
                    raise
        self.addCleanup(cleanup)

    def trait_delete(self, name):
        cmd = 'trait delete %s' % name
        self.openstack(cmd)

    def resource_provider_trait_list(self, uuid):
        cmd = 'resource provider trait list %s ' % uuid
        return self.openstack(cmd, use_json=True)

    def resource_provider_trait_set(self, uuid, *traits):
        cmd = 'resource provider trait set %s ' % uuid
        cmd += ' '.join('--trait %s' % trait for trait in traits)
        return self.openstack(cmd, use_json=True)

    def resource_provider_trait_delete(self, uuid):
        cmd = 'resource provider trait delete %s ' % uuid
        self.openstack(cmd)

    def allocation_candidate_list(self, resources=None, required=None,
                                  forbidden=None, limit=None,
                                  aggregate_uuids=None, member_of=None,
                                  may_print_to_stderr=False):
        cmd = 'allocation candidate list '
        cmd += self._allocation_candidates_option(
            resources, required, forbidden, aggregate_uuids, member_of)
        if limit is not None:
            cmd += ' --limit %d' % limit
        return self.openstack(
            cmd, use_json=True, may_print_to_stderr=may_print_to_stderr)

    def allocation_candidate_granular(self, groups, group_policy=None,
                                      limit=None):
        cmd = 'allocation candidate list '
        for suffix, req_group in groups.items():
            if suffix:
                cmd += ' --group %s' % suffix
            cmd += self._allocation_candidates_option(**req_group)
            if limit is not None:
                cmd += ' --limit %d' % limit
            if group_policy is not None:
                cmd += ' --group-policy %s' % group_policy
        return self.openstack(cmd, use_json=True)

    def _allocation_candidates_option(self, resources=None, required=None,
                                      forbidden=None, aggregate_uuids=None,
                                      member_of=None):
        opt = ''
        if resources:
            opt += ' ' + ' '.join(
                '--resource %s' % resource for resource in resources)
        if required is not None:
            opt += ''.join([' --required %s' % t for t in required])
        if forbidden:
            opt += ' ' + ' '.join('--forbidden %s' % f for f in forbidden)
        if aggregate_uuids:
            opt += ' ' + ' '.join(
                '--aggregate-uuid %s' % a for a in aggregate_uuids)
        if member_of:
            opt += ' ' + ' '.join(['--member-of %s' % m for m in member_of])
        return opt



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from osc_lib.command import command
from osc_lib import utils

from osc_placement import version


BASE_URL = '/resource_classes'
PER_CLASS_URL = BASE_URL + '/{name}'
FIELDS = ('name',)


class ListResourceClass(command.Lister):

    """Return a list of all resource classes.

    This command requires at least ``--os-placement-api-version 1.2``.
    """

    def get_parser(self, prog_name):
        parser = super(ListResourceClass, self).get_parser(prog_name)

        return parser

    @version.check(version.ge('1.2'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        resource_classes = http.request(
            'GET', BASE_URL).json()['resource_classes']
        rows = (utils.get_dict_properties(i, FIELDS) for i in resource_classes)
        return FIELDS, rows


class CreateResourceClass(command.Command):

    """Create a new resource class.

    This command requires at least ``--os-placement-api-version 1.2``.
    """

    def get_parser(self, prog_name):
        parser = super(CreateResourceClass, self).get_parser(prog_name)

        parser.add_argument(
            'name',
            metavar='<name>',
            help='Name of the resource class'
        )
        return parser

    @version.check(version.ge('1.2'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        http.request('POST', BASE_URL, json={'name': parsed_args.name})


class SetResourceClass(command.Command):

    """Create or validate the existence of single resource class.

    Unlike ``openstack resource class create``, this command also succeeds if
    the resource class already exists, which makes this an idempotent check or
    create command.

    This command requires at least ``--os-placement-api-version 1.7``.
    """

    def get_parser(self, prog_name):
        parser = super(SetResourceClass, self).get_parser(prog_name)

        parser.add_argument(
            'name',
            metavar='<name>',
            help='Name of the resource class'
        )
        return parser

    @version.check(version.ge('1.7'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = BASE_URL + '/' + parsed_args.name

        http.request('PUT', url)


class ShowResourceClass(command.ShowOne):

    """Return a representation of the resource class identified by ``<name>``.

    This command requires at least ``--os-placement-api-version 1.2``.
    """

    def get_parser(self, prog_name):
        parser = super(ShowResourceClass, self).get_parser(prog_name)

        parser.add_argument(
            'name',
            metavar='<name>',
            help='Name of the resource class'
        )

        return parser

    @version.check(version.ge('1.2'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = PER_CLASS_URL.format(name=parsed_args.name)

        resource = http.request('GET', url).json()
        return FIELDS, utils.get_dict_properties(resource, FIELDS)


class DeleteResourceClass(command.Command):

    """Delete the resource class identified by ``<name>``.

    Only custom resource classes can be deleted.

    This command requires at least ``--os-placement-api-version 1.2``.
    """

    def get_parser(self, prog_name):
        parser = super(DeleteResourceClass, self).get_parser(prog_name)

        parser.add_argument(
            'name',
            metavar='<name>',
            help='Name of the resource class'
        )

        return parser

    @version.check(version.ge('1.2'))
    def take_action(self, parsed_args):
        http = self.app.client_manager.placement

        url = PER_CLASS_URL.format(name=parsed_args.name)

        http.request('DELETE', url)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from osc_placement.tests.functional import base


CUSTOM_RC = 'CUSTOM_GPU_DEVICE_{}'.format(
    str(uuid.uuid4()).replace('-', '').upper())


class TestResourceClass(base.BaseTestCase):
    VERSION = '1.2'

    def test_list(self):
        rcs = self.resource_class_list()
        names = [rc['name'] for rc in rcs]
        self.assertIn('VCPU', names)
        self.assertIn('MEMORY_MB', names)
        self.assertIn('DISK_GB', names)

    def test_fail_create_if_incorrect_class(self):
        self.assertCommandFailed('JSON does not validate',
                                 self.resource_class_create, 'fake_class')
        self.assertCommandFailed('JSON does not validate',
                                 self.resource_class_create, 'CUSTOM_lower')
        self.assertCommandFailed('JSON does not validate',
                                 self.resource_class_create,
                                 'CUSTOM_GPU.INTEL')

    def test_create(self):
        self.resource_class_create(CUSTOM_RC)
        rcs = self.resource_class_list()
        names = [rc['name'] for rc in rcs]
        self.assertIn(CUSTOM_RC, names)
        self.resource_class_delete(CUSTOM_RC)

    def test_fail_show_if_unknown_class(self):
        self.assertCommandFailed('No such resource class',
                                 self.resource_class_show, 'UNKNOWN')

    def test_show(self):
        rc = self.resource_class_show('VCPU')
        self.assertEqual('VCPU', rc['name'])

    def test_fail_delete_unknown_class(self):
        self.assertCommandFailed('No such resource class',
                                 self.resource_class_delete, 'UNKNOWN')

    def test_fail_delete_standard_class(self):
        self.assertCommandFailed('Cannot delete standard resource class',
                                 self.resource_class_delete, 'VCPU')


class TestResourceClass17(base.BaseTestCase):
    VERSION = '1.7'

    def test_set_resource_class(self):
        self.resource_class_create(CUSTOM_RC)
        self.resource_class_set(CUSTOM_RC)
        self.resource_class_set(CUSTOM_RC + '1')
        rcs = self.resource_class_list()
        names = [rc['name'] for rc in rcs]
        self.assertIn(CUSTOM_RC, names)
        self.assertIn(CUSTOM_RC + '1', names)
        self.resource_class_delete(CUSTOM_RC)
        self.resource_class_delete(CUSTOM_RC + '1')
